{ **********************************************************************
  *                           Program SPEED.PAS                        *
  *                              Version 1.7                           *
  *                     (c) J. Debord, December 2002                   *
  **********************************************************************
  This program compares the execution times of the Pascal and Assembler
  versions of the mathematical functions provided in the unit FMATH.PAS.

  The assembler functions are selected by defining the symbol CPU387
  with a 387, 486 or Pentium processor

  The results are stored in two different output files: SPEEDASM.OUT
  if the assembler version is used, SPEED.OUT otherwise. So, the execution
  times may be compared by displaying the two files.

  Be sure to compile both the program and the unit FMATH with the same
  symbol definitions.

  NB: The execution time of each function takes into account the
      computation of random arguments. This corresponds to the
      execution time of the function Random.
  ********************************************************************** }

program Speed;

uses
  Dos, FMath;

const
  NFUNC = 27;       { Number of functions }
  NMAX  = 1000000;  { Number of evaluations for each function }

const
  FuncName : array[1..NFUNC] of String[7] =
  ('Ln     ',
   'Log10  ',
   'Log2   ',
   'Exp    ',
   'Exp10  ',
   'Exp2   ',
   'Power  ',
   'Sin    ',
   'Cos    ',
   'Tan    ',
   'ArcSin ',
   'ArcCos ',
   'ArcTan ',
   'ArcTan2',
   'Sinh   ',
   'Cosh   ',
   'Tanh   ',
   'ArcSinh',
   'ArcCosh',
   'ArcTanh',
   'Gamma  ',
   'IGamma ',
   'Beta   ',
   'IBeta  ',
   'Erf    ',
   'Erfc   ',
   'Random ');

const
  FILENAME =
{$IFDEF CPU387}
  'speedasm.out';  { Name of output file }
{$ELSE}
  'speed.out';
{$ENDIF}

var
  F : Text;   { Output file }
  N : Byte;   { Function index }
  T : Float;  { Time }

function Time : Float;
{ Returns time in seconds }
var
  H, M, S, C : Word;
begin
  GetTime(H, M, S, C);
  Time := 3600.0 * H + 60.0 * M + S + 0.01 * C;
end;

function TimeToEval(N : Byte) : Float;
{ Returns time to evaluate NMAX functions }
var
  I : LongInt;
  T0, Y : Float;
begin
  T0 := Time;
  case N of
    1 : for I := 1 to NMAX do
          Y := {$IFDEF CPU387}fLn{$ELSE}Ln{$ENDIF}(Random);
    2 : for I := 1 to NMAX do
          Y := {$IFDEF CPU387}fLog10{$ELSE}Log10{$ENDIF}(Random);
    3 : for I := 1 to NMAX do
          Y := {$IFDEF CPU387}fLog2{$ELSE}Log2{$ENDIF}(Random);
    4 : for I := 1 to NMAX do
          Y := {$IFDEF CPU387}fExp{$ELSE}Exp{$ENDIF}(Random);
    5 : for I := 1 to NMAX do
          Y := {$IFDEF CPU387}fExp10{$ELSE}Exp10{$ENDIF}(Random);
    6 : for I := 1 to NMAX do
          Y := {$IFDEF CPU387}fExp2{$ELSE}Exp2{$ENDIF}(Random);
    7 : for I := 1 to NMAX do
          Y := Power(Random, 0.5);
    8 : for I := 1 to NMAX do
          Y := {$IFDEF CPU387}fSin{$ELSE}Sin{$ENDIF}(Random);
    9 : for I := 1 to NMAX do
          Y := {$IFDEF CPU387}fCos{$ELSE}Cos{$ENDIF}(Random);
   10 : for I := 1 to NMAX do
          Y := {$IFDEF CPU387}fTan{$ELSE}Tan{$ENDIF}(Random);
   11 : for I := 1 to NMAX do
          Y := {$IFDEF CPU387}fArcSin{$ELSE}ArcSin{$ENDIF}(Random);
   12 : for I := 1 to NMAX do
          Y := {$IFDEF CPU387}fArcCos{$ELSE}ArcCos{$ENDIF}(Random);
   13 : for I := 1 to NMAX do
          Y := {$IFDEF CPU387}fArcTan{$ELSE}ArcTan{$ENDIF}(Random);
   14 : for I := 1 to NMAX do
          Y := {$IFDEF CPU387}fArcTan2{$ELSE}ArcTan2{$ENDIF}(Random, 0.5);
   15 : for I := 1 to NMAX do
          Y := Sinh(Random);
   16 : for I := 1 to NMAX do
          Y := Cosh(Random);
   17 : for I := 1 to NMAX do
          Y := Tanh(Random);
   18 : for I := 1 to NMAX do
          Y := ArcSinh(Random);
   19 : for I := 1 to NMAX do
          Y := ArcCosh(Random + 1.0);
   20 : for I := 1 to NMAX do
          Y := ArcTanh(Random);
   21 : for I := 1 to NMAX do
          Y := Gamma(Random);
   22 : for I := 1 to NMAX do
          Y := IGamma(Random, 0.5);
   23 : for I := 1 to NMAX do
          Y := Beta(Random, 0.5);
   24 : for I := 1 to NMAX do
          Y := IBeta(Random, 0.5, 0.5);
   25 : for I := 1 to NMAX do
          Y := Erf(Random);
   26 : for I := 1 to NMAX do
          Y := Erfc(Random);
   27 : for I := 1 to NMAX do
          Y := Random;
  end;
  TimeToEval := Time - T0;
end;

begin
  { Open output file }
  Assign(F, FILENAME);
  Rewrite(F);

  Writeln;
  Writeln('Time in seconds to evaluate ', NMAX, ' functions');
  Writeln(F, 'Time in seconds to evaluate ', NMAX, ' functions');
  Writeln;
  Writeln(F);

  for N := 1 to NFUNC do
    begin
      T := TimeToEval(N);
      Writeln(FuncName[N], T:6:2);
      Writeln(F, FuncName[N], T:8:2);
    end;

  Close(F);
end.
